(*
 Sunae Seo, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)


structure PpParseTree =
struct

open ParseTree
exception InvalidAst of string  

val idt = 2      // indentation size 
fun ps s = print_string s

fun pnl () = print_newline ()
fun sp n = if n<0 then raise InvalidAst "spaces"
           else (for i=0; i<n; i+1 do ps " " end)

fun p_list prfn bef aft sep elist =
  let fun list_print [e] = prfn e
       |  list_print (h::tlist) =
      prfn h; ps sep; list_print tlist
       |  list_print [] = ()
  in
    if elist = [] then ()
    else (ps bef; list_print elist; ps aft)
  end

fun p_llist prfn bef aft sep n elist =
  let fun list_print [e] = prfn e
       |  list_print (h::tlist) =
      prfn h; ps sep; pnl(); sp n; list_print tlist
       |  list_print [] = ()
  in
    if elist = [] then ()
    else (ps bef; list_print elist; ps aft)
  end

fun pp_two fn1 fn2 (v1,v2) = ps "("; fn1 v1; ps ","; fn2 v2; ps ")"

fun pp_option pf v = 
 case v of 
  Some a => ps "Some("; pf a; ps ")"
  | None => ps "None"
	   
fun pp_qualifier a = 
 case a of 
  CONST  => ps "CONST"
  | VOLATILE => ps "VOLATILE"

fun pp_storage a =
 case a of 
  TYPEDEF => ps "TYPEDEF"
  | STATIC => ps "STATIC" 
  | EXTERN  => ps "EXTERN"
  | REGISTER  => ps "REGISTER"
  | AUTO => ps "AUTO"

fun pp_operator i a =
 case a of 
  Plus => ps "Plus"
  | Minus  => ps "Minus"
  | Times  => ps "Times"
  | Divide  => ps "Divide"
  | Mod => ps "Mode"
  | Gt  => ps "Gt"
  | Lt  => ps "Lt"
  | Gte  => ps "Gte"
  | Lte  => ps "Lte"
  | Eq  => ps "Eq"
  | Neq  => ps "Neq"
  | And  => ps "And"
  | Or => ps "Or"
  | BitOr  => ps "BitOr"
  | BitAnd  => ps "BitAnd"
  | BitXor  => ps "BitXor"
  | Lshift  => ps "Lshift"
  | Rshift => ps "Rshift"
  | Star  => ps "Star"
  | AddrOf  => ps "AddrOf"
  | Dot  => ps "Dot"
  | Arrow  => ps "Arrow"
  | Sub  => ps "Sub"
  | Sizeof => ps "Sizeof"
  | PreInc  => ps "PreInc"
  | PostInc  => ps "PostInc"
  | PreDec  => ps "PreDec"
  | PostDec  => ps "PostDec"
  | Comma => ps "Comma"
  | Not  => ps "Not"
  | Negate  => ps "Negate"
  | BitNot  => ps "BitNot"
  | Assign => ps "Assign"
  | PlusAssign  => ps "PlusAssign"
  | MinusAssign  => ps "MinusAssign"
  | TimesAssign  => ps "TimesAssign"
  | DivAssign => ps "DivAssign"
  | ModAssign  => ps "ModAssign"
  | XorAssign  => ps "XorAssign"
  | OrAssign  => ps "OrAssign"
  | AndAssign => ps "AndAssign"
  | LshiftAssign  => ps "LshiftAssign"
  | RshiftAssign  => ps "RshiftAssign"
  | Uplus  => ps "Uplus"
  | SizeofType ct => ps "SizeofType "; pp_ctype i ct

and pp_expression i a =
 case a of
  EmptyExpr => ps "EmptyExpr"
  | IntConst v => ps "IntConst "; print_int v
  | RealConst v => ps "RealConst "; ps v
  | String s => ps "String "; ps s 
  | Id s => ps "Id "; ps s
  | Unop  (op, e) => ps "Unop("; pp_operator i op; pp_expression i e; ps ")"
  | Binop (op, e1, e2) => 
     ps "Binop(";
     pp_operator i op; ps "("; 
     pp_expression i e1; ps ","; 
     pp_expression i e2; ps "))"
  | QuestionColon (e1, e2, e3) => 
     ps "QuestionColon(";
     pp_expression i e1; ps ",";
     pp_expression i e2; ps ",";
     pp_expression i e3; ps "))"
  | Call (e1, e2) => 
     ps "Call(";
     pp_expression i e1; ps ",";
     p_list (pp_expression i) "[" "]" "," e2; ps "))"
  | Cast (ct, e) =>
     ps "Cast(";
     pp_ctype i ct; ps ",";
     pp_expression i e; ps ")"
  | InitList elist => 
     ps "InitList"; 
     p_list (pp_expression i) "[" "]" "," elist
  | MARKexpression e => ps "MARKexpression "; pp_expression i e 

and pp_specifier i a =
 case a of 
  Void => ps "Void"
  | Ellipses => ps "Ellipses"
  | Signed => ps "Signed"
  | Unsigned => ps "Unsigned"
  | Char => ps "Char"
  | Short => ps "Short"
  | Int => ps "Int"
  | Long => ps "Long"
  | Float  => ps "Float"
  | Double => ps "Double"
  | Fractional => ps "Fractional"
  | Wholenum => ps "Wholenum"
  | Saturate => ps "Saturate"
  | Nonsaturate => ps "Nonsaturate"
  | Array (e, ct) => 
     ps "Array(";
     pp_expression i e; ps ",";
     pp_ctype i ct; ps ")"
  | Pointer ct => ps "Pointer"; pp_ctype i ct
  | Function {retType, params} => 
     ps "Function(";
     pp_ctype i retType;
     p_list (pp_two (pp_decltype i) (pp_declarator i)) "[" "]" "," params; 
     ps ")"
  | Enum {tagOpt, enumerators, trailingComma}  => 
     ps "Enum("; 
     pp_option ps tagOpt; ps ",";
     p_list (pp_two ps (pp_expression i))  "[" "]" "," enumerators; 
     ps ",";
     Format.print_bool trailingComma; ps ")"
  | Struct {isStruct, tagOpt, members} => 
     ps "Struct(";
     Format.print_bool isStruct; ps ",";
     pp_option ps tagOpt; ps ",";
     let fun my_pptwo (ct,delist) = 
         ps "(" ; pp_ctype i ct; ps ",";
 	 p_list (pp_two (pp_declarator i) (pp_expression i)) "[" "]" "," delist;
 	 ps ")"
     in
        p_list my_pptwo "[" "]" "," members
     end; ps ")"
  | TypedefName str => 
     ps "TypedefName "; ps str
  | StructTag {isStruct, name} => 
     ps "StructTag(" ; 
     Format.print_bool isStruct; ps ",";
     ps name ; ps ")"
  | EnumTag str  => ps "EnumTag " ; ps str

and pp_declarator i a =
 case a of 
  EmptyDecr => ps "EmptyDecr"
  | EllipsesDecr => ps "EllipsesDecr"
  | VarDecr str => ps "VarDecr("; ps str; ps ")"
  | ArrayDecr (dec, e) => 
     ps "EmptyDecr(";
     pp_declarator i dec; ps ",";
     pp_expression i e; ps ")"
  | PointerDecr dec => ps "PointerDecr("; pp_declarator i dec; ps ")"
  | QualDecr (q, d) => 
     ps "QualDecr(";
     pp_qualifier q; ps ",";
     pp_declarator i d; ps ")"
  | FuncDecr (d, ddlist) => 
     ps "FuncDecr(";
     pp_declarator i d; ps ",";
     p_list (pp_two (pp_decltype i) (pp_declarator i)) "[" "]" "," ddlist; 
     ps ")"
  | MARKdeclarator d => 
     ps "MARKdeclarator("; pp_declarator i d; ps ")"

and pp_statement i a =
 case a of 
  Decl d => ps "Decl(\n"; 
    sp i; pp_declaration (i+idt) d; ps "\n";
    sp i; ps ")"
  | Expr e => ps "Expr("; pp_expression (i+idt) e; ps ")"
  | Compound stlist => 
     ps "Compound[\n"; 
     sp i; p_llist (pp_statement (i+idt)) "" "\n" "," (i+idt) stlist;
     sp i; ps "]" 
  | While (e,s) =>
     ps "While(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | Do (e,s) =>
     ps "Do(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | For (e1,e2,e3,s) =>
     ps "For(\n";
     sp i; pp_expression (i+idt) e1; ps ",\n";
     sp i; pp_expression (i+idt) e2; ps ",\n";
     sp i; pp_expression (i+idt) e3; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | Labeled (str,s) =>
     ps "Labeled(\n";
     sp i; ps str; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | CaseLabel (e,s) =>
     ps "CaseLabel(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | DefaultLabel s => 
     ps "DefaultLabel(\n"; 
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | Goto str => ps "Goto("; ps str; ps ")"
  | Break => ps "Break"
  | Continue => ps "Continue"
  | Return e => ps "Return("; pp_expression (i+idt) e; ps ")"
  | IfThen (e,s) => 
     ps "IfThen(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | IfThenElse (e,s1,s2) =>
     ps "IfThenElse(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s1; ps ",\n";
     sp i; pp_statement (i+idt) s1; ps "\n";
     sp i; ps ")"
  | Switch (e,s) =>
     ps "Switch(\n";
     sp i; pp_expression (i+idt) e; ps ",\n";
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"
  | MARKstatement s => 
     ps "MARKstatement(\n"; 
     sp i; pp_statement (i+idt) s; ps "\n";
     sp i; ps ")"

and pp_declaration i a =
 case a of 
  Declaration (dt, delist) =>
     ps "Declaration(\n";
     sp i; pp_decltype (i+idt) dt; ps ",\n";
     sp i; p_list (pp_two (pp_declarator (i+idt)) (pp_expression (i+idt))) "[" "]" "," delist; ps "\n";
     sp i; ps ")"
  | MARKdeclaration d => 
     ps "MARKdeclaration(\n"; 
     sp i; pp_declaration (i+idt) d; ps "\n";
     sp i; ps ")"

and pp_externalDecl i a =
 case a of 
  ExternalDecl d => ps "ExternalDecl(\n"; 
     sp i; pp_declaration (i+idt) d; ps "\n";
     sp i; ps ")"
  | FunctionDef {retType,funDecr,krParams,body} =>
     ps "FunctionDef{\n";
     sp i; ps "retType:"; pp_decltype (i+idt) retType; ps ",\n";
     sp i; ps "funDecr:"; pp_declarator (i+idt) funDecr; ps ",\n";
     sp i; ps "krParams:"; 
        p_llist (pp_declaration (i+idt)) "[" "]" "," (i+idt) krParams; ps ",\n";
     sp i; ps "body:\n"; 
     sp i; pp_statement (i+idt) body; 
     sp i; ps "}"
  | MARKexternalDecl ex =>
     ps "MARKexternalDecl(\n"; 
     sp i; pp_externalDecl (i+idt) ex; ps "\n";
     sp i; ps ")" 

and pp_top a = p_llist (pp_externalDecl 0) "" "" "," 0 a

and pp_ctype i {qualifiers, specifiers} =
     ps "{qualifiers:"; p_list pp_qualifier "[" "]" "," qualifiers; ps ",\n";
     sp i; ps "specifiers:"; 
       p_list (pp_specifier (i+idt)) "[" "]" "," specifiers; ps "}"
and pp_decltype i {qualifiers, specifiers,storage} =
     ps "{qualifiers:"; p_list pp_qualifier "[" "]" "," qualifiers; ps ",\n";
     sp i; ps "specifiers:"; 
       p_list (pp_specifier (i+idt)) "[" "]" "," specifiers; ps ",\n";
     sp i; ps "storage:"; p_list pp_storage "[" "]" "," storage; ps "}"

end (* structure ParseTree *)
